from pathlib import Path

from .quarto_render import quarto_render_html
from .chrome_print import chrome_print
from .util.util import filename_with_suffix

from .pdf_to_pptx import pdf_to_pptx


def qmd_to_pdf(
    input_filename: str,
    output_filename: str = None,
    delete_intermediate: bool = True,
    browser: str = None,
    page_load_time: int = 1,
    verbose: int = 0,

):
    """
    Renders a Quarto (qmd) revealjs presentation and converts this to a pdf file.

    pptx is generated via rendering the presentation in html and printing html to pdf.

    Args:
        input_filename (str): Path to the input qmd file.
        output_filename (str): Path to save the output PDF file.
        delete_intermediate (bool, optional): If True, delete the html outputs after pdf is generated. Default is True.
        browser (str, optional): The path to chrome (or chrome based brower like chromium / edge). Defaults to the default location for the OS.
        page_load_time (int, optional): Adds a wait time before attempting to print the file to allow the page to fully load. Defaults to 1 second. If output is a blank page increase this time.
        verbose (bool, optional): If True, show progress messages. Default is True.
        
    Retuns:
        None

    Examples:
        >>> qmd_to_pptx('document.qmd', 'output.pdf')
    """

    if output_filename is None:
        output_filename = filename_with_suffix(input_filename, "pdf")

    filename_qmd = filename_with_suffix(input_filename, "qmd")
    filename_html = filename_with_suffix(output_filename, "html")
    filename_pdf = filename_with_suffix(output_filename, "pdf")

    quarto_render_html(
        input_filename=filename_qmd, output_filename=filename_html, verbose=verbose
    )

    chrome_print(
        input_file=filename_html,
        output_file=filename_pdf,
        browser=browser,
        verbose=verbose,
        page_load_time=page_load_time
    )

    if delete_intermediate:
        Path(filename_html).unlink()


def qmd_to_pptx(
    input_filename: str,
    output_filename: str = None,
    delete_intermediate: bool = True,
    browser: str = None,
    dpi: int = 300,
    pptx_height: float = 7.5,
    pptx_width: float = 13.33,
    page_load_time: int = 1,
    verbose: int = 0,
):
    """
    Renders a Quarto (qmd) revealjs presentation and converts this to a pptx file.

    pptx is generated via rendering the presentation in html, printing html to pdf and then taking a screenshot of each page and inserting into a blank pptx.

    Args:
        input_filename (str): Path to the input qmd file.
        output_filename (str): Path to save the output PowerPoint presentation.
        delete_intermediate (bool, optional): If True, delete both the html and pdf outputs after pptx is generated. Default is True.
        browser (str, optional): The path to chrome (or chrome based brower like chromium / edge). Defaults to the default location for the OS.
        dpi (int, optional): DPI (dots per inch) for image conversion. Default is 300.
        pptx_height(float, optional): PPTX height in inches. Default is 7.5 (standard widescreen presentation size)
        pptx_width(float, optional): PPTX width in inches. Default is 13.33 (standard widescreen presentation size)
        page_load_time (int, optional): Adds a wait time before attempting to print the file to allow the page to fully load. Defaults to 1 second. If output is a blank page increase this time.
        verbose (bool, optional): If True, show progress messages. Default is True.
        
    Retuns:
        None

    Examples:
        >>> qmd_to_pptx('document.qmd', 'output.pptx', dpi=500)
    """

    if output_filename is None:
        output_filename = filename_with_suffix(input_filename, "pptx")

    filename_qmd = filename_with_suffix(input_filename, "qmd")
    filename_html = filename_with_suffix(output_filename, "html")
    filename_pdf = filename_with_suffix(output_filename, "pdf")
    filename_pptx = filename_with_suffix(output_filename, "pptx")

    quarto_render_html(
        input_filename=filename_qmd, output_filename=filename_html, verbose=verbose
    )

    chrome_print(
        input_file=filename_html,
        output_file=filename_pdf,
        browser=browser,
        verbose=verbose,
        page_load_time=page_load_time
    )

    pdf_to_pptx(
        input_pdf_path=filename_pdf,
        output_pptx_path=filename_pptx,
        dpi=dpi,
        pptx_height=pptx_height,
        pptx_width=pptx_width,
        verbose=verbose,
    )

    if delete_intermediate:
        Path(filename_html).unlink()
        Path(filename_pdf).unlink()



from pathlib import Path
import subprocess as sp

from .util.util import sanitize_input


def quarto_render_html(input_filename: str, output_filename: str, verbose: int = 0):
    """
    Renders a quarto (qmd) file.

    Args:
        input_filename (str): The name of the file to be rendered.
        output_filename (str): The name of the output file.
        verbose (bool, optional): If True, prints detailed status messages. Default is False.

    Returns:
        None

    Examples:
        >>> quarto_render_html('example.qmd', 'output.html')
    """

    # if file doesn't end with .qmd we want to add this here
    if input_filename[-4:] != ".qmd":
        input_filename = input_filename + ".qmd"

    # sanitise filenames as will be used in command line
    input_filename = sanitize_input(input_filename)
    output_filename = sanitize_input(output_filename)

    # check if the file exists if not then quit the function
    if Path(input_filename).is_file() == False:
        raise Exception(
            "input_filename '" + input_filename + "' does not reference a qmd file"
        )

    sp.run(
        ["quarto", "render", input_filename, "--output=" + output_filename],
        capture_output=False if verbose >= 1 else True,
    )

    # check if the output html file was created if not then quit the function
    if Path(output_filename).is_file() == False:
        raise Exception("failed to generate html file from '" + input_filename + "'")



import http.server
import socketserver
import threading
import os


class MyHttpRequestHandler(http.server.SimpleHTTPRequestHandler):
    html_file = None
    base_dir = None

    def end_headers(self):
        self.send_header("Cache-Control", "no-store, must-revalidate")
        http.server.SimpleHTTPRequestHandler.end_headers(self)

    def translate_path(self, path):
        # Serve the requested file if it exists, otherwise serve the HTML file
        requested_path = os.path.join(MyHttpRequestHandler.base_dir, path.lstrip("/"))
        if os.path.exists(requested_path) and not os.path.isdir(requested_path):
            return requested_path
        else:
            return os.path.join(MyHttpRequestHandler.base_dir, MyHttpRequestHandler.html_file)

    def log_message(self, format, *args):
        self.server.log(f"Server: {format % args}")


class LocalWebServer:
    def __init__(self, html_file, port=8000, verbose=0):
        MyHttpRequestHandler.html_file = os.path.basename(html_file)
        MyHttpRequestHandler.base_dir = os.path.dirname(os.path.abspath(html_file))
        self.port = port
        self.verbose = verbose
        self.server_thread = None
        self.httpd = None

    def _log(self, message):
        if self.verbose >= 1:
            print(message)

    def start(self):
        if not self.server_thread or not self.server_thread.is_alive():
            self.httpd = socketserver.TCPServer(("", self.port), MyHttpRequestHandler)
            self.httpd.log = self._log  # Assign _log method to server.log
            self.server_thread = threading.Thread(target=self.httpd.serve_forever)
            self.server_thread.daemon = True
            self.server_thread.start()
            self._log(
                f"Server started at port {self.port}. Serving {MyHttpRequestHandler.html_file}"
            )

    def close(self):
        if self.server_thread and self.server_thread.is_alive():
            self.httpd.shutdown()
            self.httpd.server_close()
            self.server_thread.join()
            self._log("Server closed.")


import os
from pathlib import Path
from dotenv import load_dotenv
import shutil
import socket
import datetime
import subprocess
import time
import requests

from .util.util import filename_with_suffix
from .print_pdf.print_url import print_url
from .print_pdf.serve_html_file import LocalWebServer

import platform

if platform.system() == "Windows":
    import winreg


def chrome_print(
    input_file: str, output_file: str = None, browser: str = None, verbose: int = 0, page_load_time:int=1
):
    """
    Prints a local html file to PDF through chrome.
    Designed print the output of a rendered quarto file.

    Args:
        input_filename (str): The name of the html file to be printed via chrome.
        output_filename (str, optional): The name of the output PDF file. Defaults to the same name as the input file.
        browser (str, optional): The path to chrome (or chrome based brower like chromium / edge). Defaults to the default location for the OS.
        page_load_time (int, optional): Adds a wait time before attempting to print the file to allow the page to fully load. Defaults to 1 second. If output is a blank page increase this time.
        verbose (bool, optional): If True, prints detailed status messages. Default is False.
        
    Returns:
        None

    Examples:
        >>> chrome_print('document.html')
        >>> chrome_print('document.html', 'output.html')

    """

    try:
        ps = None  # this is in case it errors before ps is created. if this happens the on close statement doesn't work correctly.
        load_dotenv()  # load variables from .env if they exist

        # check if input file exists and is an html file
        if not Path(input_file).exists():
            raise Exception("Input file doesn't exist")
        if input_file[-5:] != ".html":
            raise Exception("Input file is not a html file")

        input_file = str(Path(input_file).resolve())

        # create output file from input file
        if output_file is None:
            output_file = filename_with_suffix(input_file, "pdf")

        # get chrome browser
        if browser is None:
            if os.getenv("QUARTO_CHROME") is not None:
                browser = browser = os.getenv("QUARTO_CHROME")
            else:
                browser = find_chrome()
        else:
            if Path(browser).exists() == False:
                browser = shutil.which(browser)
        if os.access(browser, os.X_OK) == False:
            raise Exception("The browser is not executable: " + browser)
        if verbose >= 1:
            print("Using browser: " + browser)

        # get a free port on your system
        debug_port = find_free_port()

        # create list of args to be passed into chrome print
        extra_args = []
        extra_args.extend(proxy_args())
        if platform.system() == "Windows":
            extra_args.append("--no-sandbox")
        extra_args.extend(
            [
                "--headless",
                "--disable-dev-shm-usage",
                "--font-render-hinting=none",
                "--remote-allow-origins=*",
                "--disable-gpu",
                "--disable-translate",
                "--disable-extensions",
                "--disable-background-networking",
                "--safebrowsing-disable-auto-update",
                "--disable-sync",
                "--disable-default-apps",
                "--hide-scrollbars",
                "--metrics-recording-only",
                "--mute-audio",
                "--no-first-run",
                "--no-default-browser-check",
            ]
        )

        extra_args = list(set(extra_args))
        extra_args.extend(
            [
                f"--remote-debugging-port={debug_port}",
            ]
        )

        log_file = (
            f"chrome-stderr-{datetime.now().strftime('%Y-%m-%d_%H-%M-%S')}.log"
            if os.getenv("pagedown.chrome.log", "FALSE").upper() == "TRUE"
            else None
        )

        ps = subprocess.Popen(
            [browser] + extra_args,
            stderr=(
                None
                if verbose >= 1
                else subprocess.DEVNULL if log_file is None else open(log_file, "w")
            ),
            stdout=None if verbose >= 1 else subprocess.DEVNULL,
        )

        remote_protocol_ok = is_remote_protocol_ok(debug_port, verbose=verbose)

        if remote_protocol_ok == False:
            raise Exception("A more recent version of Chrome is required.")

        websocket_url = get_entrypoint(debug_port, verbose)

        serve_html_port = find_free_port()
        web_server = LocalWebServer(
            input_file, port=int(serve_html_port), verbose=verbose
        )
        web_server.start()

        print_url(
            websocket_url,
            f"http://127.0.0.1:{str(serve_html_port)}?print-pdf",
            output_file,
        )

    finally:
        kill_chrome(ps, web_server, verbose)


# supporting functions


def find_chrome() -> str:
    match platform.system():
        case "Darwin":  # macos
            return "/Applications/Google Chrome.app/Contents/MacOS/Google Chrome"
        case "Linux":
            for browser in [
                "google-chrome",
                "chromium-browser",
                "chromium",
                "google-chrome-stable",
            ]:
                result = shutil.which(browser)
                if result != None:
                    return result
            # if gets to here it couldn't find anything
            raise Exception(
                "can't find google chrome specify the path in the browser variable"
            )
        case "Windows":
            for browser in ["ChromeHTML", "MSEdgeHTM"]:
                try:
                    result = winreg.QueryValueEx(
                        winreg.OpenKey(
                            winreg.HKEY_CLASSES_ROOT,
                            browser + "\\shell\\open\\command",
                            0,
                            winreg.KEY_READ,
                        ),
                        "",
                    )[0].split('"')[1]
                except:
                    result = ""
                if result != "":
                    return result
            raise Exception(
                "can't find google chrome specify the path in the browser variable"
            )
        case _:
            raise Exception("your os is not currently supported")


def proxy_args() -> str:
    # Get proxy environment variables in order of priority
    proxy_vars = ["https_proxy", "HTTPS_PROXY", "http_proxy", "HTTP_PROXY"]
    val = [os.getenv(var) for var in proxy_vars if os.getenv(var)]

    # If no non-empty proxy environment variable is found, return an empty list
    if not val:
        return []

    # Construct proxy arguments
    proxy_server_arg = f"--proxy-server={val[0]}"
    proxy_bypass_list_arg = "--proxy-bypass-list=" + ";".join(no_proxy_urls())

    return [proxy_server_arg, proxy_bypass_list_arg]


def no_proxy_urls() -> str:
    # Get no_proxy environment variable, split by ',' or ';' and include localhost and 127.0.0.1
    no_proxy_env = os.getenv("no_proxy") or os.getenv("NO_PROXY") or ""
    urls = [
        url.strip() for url in no_proxy_env.replace(",", ";").split(";") if url.strip()
    ]
    urls.extend(["localhost", "127.0.0.1"])
    return list(set(urls))  # Return unique URLs


def find_free_port() -> str:
    with socket.socket() as s:
        s.bind(("", 0))  # Bind to a free port provided by the host.
        return str(s.getsockname()[1])  # Return the port number assigned.


def kill_chrome(ps, web_server, verbose: int = 0):
    if verbose >= 1:
        print("Closing browser")

    # Check if the process is alive before attempting to kill it
    if ps is not None:
        if ps.poll() is None:
            ps.kill()

    if verbose >= 1:
        print("Closing webserver")

    if web_server is not None:
        web_server.close()


def is_remote_protocol_ok(debug_port: str, verbose: int = 0) -> bool:
    url = f"http://127.0.0.1:{debug_port}/json/protocol"

    # Specify the maximum number of attempts and sleep time between attempts
    max_attempts = 20
    sleep_time = 0.5

    if verbose >= 1:
        print(f"Trying to find headless Chrome in {max_attempts} attempts")

    for i in range(1, max_attempts + 1):
        try:
            response = requests.get(url)
            response.raise_for_status()  # Raise an error for bad response status codes

            remote_protocol = response.json()
        except Exception as e:
            remote_protocol = None

        if remote_protocol:
            if verbose >= 1:
                print(f"Headless Chrome found at attempt {i}")
            break

        if i == max_attempts:
            raise RuntimeError(
                f"Cannot find headless Chrome after {max_attempts} attempts"
            )

        time.sleep(sleep_time)

    # Define required commands and events
    required_commands = {
        "DOM": ["enable", "getBoxModel", "getDocument", "querySelector"],
        "Network": ["enable"],
        "Page": [
            "addScriptToEvaluateOnNewDocument",
            "captureScreenshot",
            "enable",
            "navigate",
            "printToPDF",
        ],
        "Runtime": ["enable", "addBinding", "evaluate"],
        "Target": ["attachToTarget", "createTarget"],
    }

    required_events = {
        "Inspector": ["targetCrashed"],
        "Network": ["responseReceived"],
        "Page": ["loadEventFired"],
        "Runtime": ["bindingCalled", "exceptionThrown"],
    }

    # Extract remote domains, commands, and events
    remote_domains = [domain["domain"] for domain in remote_protocol["domains"]]
    remote_commands = {
        domain: [
            command["name"]
            for d in remote_protocol["domains"]
            if d["domain"] == domain
            for command in d["commands"]
        ]
        for domain in required_commands.keys()
    }
    remote_events = {
        domain: [
            event["name"]
            for d in remote_protocol["domains"]
            if d["domain"] == domain
            for event in d["events"]
        ]
        for domain in required_events.keys()
    }

    # Check if all required commands and events are available in the remote protocol
    if not all(
        set(required_commands[domain]) <= set(remote_commands[domain])
        for domain in required_commands
    ):
        return False
    if not all(
        set(required_events[domain]) <= set(remote_events[domain])
        for domain in required_events
    ):
        return False

    # Check if printToPDF supports streaming
    print_to_pdf_params = [
        param
        for param in remote_protocol["domains"][remote_domains.index("Page")][
            "commands"
        ]
        if param["name"] == "printToPDF"
    ][0]["parameters"]
    stream_pdf_available = "transferMode" in [
        param["name"] for param in print_to_pdf_params
    ]

    return stream_pdf_available


def get_entrypoint(debug_port: str, verbose: int = 0) -> str:
    url = f"http://127.0.0.1:{debug_port}/json/version"
    response = requests.get(url)
    version_infos = response.json()

    browser = version_infos.get("webSocketDebuggerUrl")
    if not browser:
        raise Exception("Cannot find 'Browser' websocket URL. Please retry.")

    if verbose >= 1:
        print("Browser version:", version_infos.get("Browser"))

    return browser



from pathlib import Path
from re import sub
from shlex import quote
import os


def filename_with_suffix(filename: str, suffix: str) -> str:
    return Path(filename).stem + "." + suffix


def sanitize_input(user_input: str) -> str:
    if os.name == "posix":
        # Sanitize for POSIX command line
        sanitized_input = quote(user_input)
        # Remove invalid filename characters for POSIX (if any)
        sanitized_input = sub(r"[:]", "", sanitized_input)
    elif os.name == "nt":
        # Define a regex pattern for potentially dangerous characters for Windows CMD
        dangerous_chars = r"[&|;^<>()%]"
        # Replace dangerous characters with an empty string
        sanitized_input = sub(dangerous_chars, "", user_input)
        # Enclose the sanitized input in double quotes
        sanitized_input = f'"{sanitized_input}"'
        # Define a regex pattern for invalid filename characters on Windows
        invalid_filename_chars = r'[\\/:"*?<>|]'
        # Remove invalid filename characters
        sanitized_input = sub(invalid_filename_chars, "", sanitized_input)
    else:
        raise NotImplementedError("Unsupported operating system")

    return sanitized_input



from .quarto_output import qmd_to_pdf, qmd_to_pptx
from .quarto_render import quarto_render_html
from .chrome_print import chrome_print
from .pdf_to_pptx import pdf_to_pptx

__all__ = [
    "qmd_to_pdf",
    "qmd_to_pptx",
    "quarto_render_html",
    "chrome_print",
    "pdf_to_pptx",
]



from ..quarto_output import qmd_to_pdf, qmd_to_pptx

import argparse


def _cli():
    """
    Function for running qmd_to_pdf or qmd_to_pptx via command line.
    Use either of the above functions instead if running via a python script.
    """

    parser = argparse.ArgumentParser(
        description="Create a PDF or PPTX from your quarto template."
    )

    parser.add_argument("input", type=str, help="The input QMD file name")
    parser.add_argument(
        "-t",
        "--type",
        type=str,
        default="pdf",
        help="The output type (pdf or pptx). (optional, defaults to pdf)",
    )
    parser.add_argument(
        "-o",
        "--output",
        type=str,
        default=None,
        help="The output PDF/PPTX filename (optional)",
    )
    parser.add_argument(
        "-di",
        "--delete_intermediates",
        action="store_false",
        help="Do not delete intermediate files if included (optional)",
    )
    parser.add_argument(
        "-b",
        "--browser",
        type=str,
        default=None,
        help="Specify the browser to use (optional)",
    )
    parser.add_argument(
        "--dpi",
        type=int,
        default=300,
        help="DPI for the PPTX slides (optional, defaults to 300)",
    )
    parser.add_argument(
        "-ph",
        "--pptx_height",
        type=float,
        default=7.5,
        help="Height of the PPTX slide in inches (optional, defaults to 7.5)",
    )
    parser.add_argument(
        "-pw",
        "--pptx_width",
        type=float,
        default=13.33,
        help="Width of the PPTX slide in inches (optional, defaults to 13.33)",
    )
    parser.add_argument(
        "-v", "--verbose", type=int, default=0, help="Verbose level (optional)"
    )
    parser.add_argument(
        "-pl", "--page_load", type=int, default=1, help="Page load time for the file in seconds. If you end up with a blank page increase this. (optional, defaults to 1)"
    )

    args = parser.parse_args()

    if args.type.lower() == "pptx":
        qmd_to_pptx(
            input_filename=args.input,
            output_filename=args.output,
            delete_intermediate=args.delete_intermediates,
            dpi=args.dpi,
            pptx_height=args.pptx_height,
            pptx_width=args.pptx_width,
            browser=args.browser,
            page_load_time=args.page_load,
            verbose=args.verbose,
        )
    else:
        qmd_to_pdf(
            input_filename=args.input,
            output_filename=args.output,
            delete_intermediate=args.delete_intermediates,
            browser=args.browser,
            page_load_time=args.page_load,
            verbose=args.verbose,
        )


if __name__ == "__main__":
    _cli()



from pdf2image import convert_from_path
from pptx import Presentation
from pptx.util import Inches
import io


def pdf_to_pptx(
    input_pdf_path: str,
    output_pptx_path: str,
    dpi: int = 300,
    pptx_height: float = 7.5,
    pptx_width: float = 13.33,
    verbose: int = 0,
):
    """
    Convert a PDF file to a PowerPoint presentation.

    Args:
        input_pdf_path (str): Path to the input PDF file.
        output_pptx_path (str): Path to save the output PowerPoint presentation.
        dpi (int, optional): DPI (dots per inch) for image conversion. Default is 300.
        pptx_height(float, optional): PPTX height in inches. Default is 7.5 (standard widescreen presentation size)
        pptx_width(float, optional): PPTX width in inches. Default is 13.33 (standard widescreen presentation size)
        verbose (bool, optional): If True, show progress messages. Default is True.

    Retuns:
        None

    Examples:
        >>> pdf_to_pptx('document.pdf', 'output.pptx', dpi=500)
    """

    # Convert PDF pages to images
    images = convert_from_path(input_pdf_path, dpi=dpi)

    # Create a PowerPoint presentation object
    presentation = Presentation()

    # Set slide dimensions to 16:9
    presentation.slide_width = Inches(pptx_width)
    presentation.slide_height = Inches(pptx_height)

    if verbose >= 1:
        print(f"Converting PDF to PPTX. Total pages: {len(images)}")

    for i, image in enumerate(images, start=1):
        # Save image to a byte array
        img_byte_arr = io.BytesIO()
        image.save(img_byte_arr, format="PNG")
        img_byte_arr.seek(0)

        # Add a slide to the presentation
        slide = presentation.slides.add_slide(
            presentation.slide_layouts[5]
        )  # Use blank layout

        # Add the image to the slide
        left = top = Inches(0)
        slide.shapes.add_picture(
            img_byte_arr,
            left,
            top,
            width=Inches(pptx_width),
            height=Inches(pptx_height),
        )  # Adjust as needed

        if verbose >= 1:
            print(f"Slide {i}/{len(images)} added to the presentation")

    # Save the PowerPoint presentation
    presentation.save(output_pptx_path)

    if verbose >= 1:
        print(f"PDF has been converted to PPTX and saved at {output_pptx_path}")



import websocket
import json
import base64
import time

print_args = {
    # "landscape": False,
    "displayHeaderFooter": False,
    "printBackground": True,
    # "paperWidth": 8.27,  # inch
    # "paperHeight": 11.69,  # inch
    "marginTop": 0,  # 1cm
    "marginBottom": 0,
    "marginLeft": 0,
    "marginRight": 0,
    "pageRanges": "",
    "ignoreInvalidPageRanges": False,
    "headerTemplate": "",
    "footerTemplate": "",
    "preferCSSPageSize": True,
}


def print_url(
    ws_url,
    url,
    output_file,
    ws_timeout=120,
    command_timeout=30,
    page_load_time=1
):
    try:

        ws = websocket.create_connection(ws_url, suppress_origin=False)
        ws.settimeout(ws_timeout)

        # 1 Target.createTarget (tab)
        ws.send(
            json.dumps(
                {
                    "id": 1,
                    "method": "Target.createTarget",
                    "params": {"url": "about:blank"},
                }
            )
        )
        result = wait_result(ws, 1, command_timeout)

        # 2 Target.attachToTarget in flat mode
        ws.send(
            json.dumps(
                {
                    "id": 2,
                    "method": "Target.attachToTarget",
                    "params": {
                        "targetId": result["result"]["targetId"],
                        "flatten": True,
                    },
                }
            )
        )
        result = wait_result(ws, 2, command_timeout)
        session_id = result["result"]["sessionId"]

        # 3 Runtime.enable
        ws.send(
            json.dumps({"id": 3, "sessionId": session_id, "method": "Runtime.enable"})
        )
        result = wait_result(ws, 3, command_timeout)

        # 4 Page.enable
        ws.send(json.dumps({"id": 4, "sessionId": session_id, "method": "Page.enable"}))
        result = wait_result(ws, 4, command_timeout)

        # 5 Runtime.addBinding
        ws.send(
            json.dumps(
                {
                    "id": 5,
                    "sessionId": session_id,
                    "method": "Runtime.addBinding",
                    "params": {"name": "pagedownListener"},
                }
            )
        )
        result = wait_result(ws, 5, command_timeout)

        # 6 Network.enable
        ws.send(
            json.dumps({"id": 6, "sessionId": session_id, "method": "Network.enable"})
        )
        result = wait_result(ws, 6, command_timeout)

        # 7 Page.navigate
        ws.send(
            json.dumps(
                {
                    "id": 7,
                    "sessionId": session_id,
                    "method": "Page.navigate",
                    "params": {"url": url},
                }
            )
        )
        result = wait_result(ws, 7, command_timeout)

        error_on_page_nav = (
            True if "result" in result and "error" in result["result"] else False
        )
        if not error_on_page_nav:
            error_on_page_nav = True if "error" in result else False
        if error_on_page_nav:
            raise Exception("Can't nagivate to " + url)

        # wait for page to load
        result = wait_event(ws, "Page.frameStoppedLoading", command_timeout)
        
        time.sleep(page_load_time)

        # 8 Page.printToPDF
        ws.send(
            json.dumps(
                {
                    "id": 8,
                    "sessionId": session_id,
                    "method": "Page.printToPDF",
                    "params": print_args,
                }
            )
        )
        result = wait_result(ws, 8, command_timeout)

        with open(output_file, "wb") as f:
            f.write(base64.decodebytes(str.encode(result["result"]["data"])))

    finally:
        ws.close()


def wait_event(ws, event, timeout=120):
    start_time = time.time()
    messages = []
    matching_message = None
    while True:
        now = time.time()
        if now - start_time > timeout:
            break
        try:
            message = ws.recv()
            parsed_message = json.loads(message)
            messages.append(parsed_message)
            if "method" in parsed_message and parsed_message["method"] == event:
                matching_message = parsed_message
                break
        except websocket.WebSocketTimeoutException:
            continue
        except Exception:
            break
    return matching_message

def wait_event(ws, event, timeout=120, sleep_interval=0.5):
    start_time = time.time()
    messages = []
    matching_message = None
    while True:
        now = time.time()
        if now - start_time > timeout:
            break
        try:
            message = ws.recv()
            parsed_message = json.loads(message)
            messages.append(parsed_message)
            if "method" in parsed_message and parsed_message["method"] == event:
                matching_message = parsed_message
                break
        except websocket.WebSocketTimeoutException:
            time.sleep(sleep_interval)
            continue
        except Exception as e:
            print(f"An error occurred: {e}")
            break
    return matching_message


def wait_result(ws, result_id, timeout=30):
    start_time = time.time()
    messages = []
    matching_result = None
    while True:
        now = time.time()
        if now - start_time > timeout:
            break
        try:
            message = ws.recv()
            parsed_message = json.loads(message)
            messages.append(parsed_message)
            if (
                "result" in parsed_message or "error" in parsed_message
            ) and parsed_message["id"] == result_id:
                matching_result = parsed_message
                break
        except websocket.WebSocketTimeoutException:
            continue
        except Exception:
            break
    return matching_result
